from enum import Enum
from typing import Any, Type

from solidlab_perftest_common.agent import (
    AgentCommand,
    Agent,
    AgentCommandResult,
)


class SolidLabContentType(Enum):
    AGENT_COMMAND = "application/vnd.imecilabt.solidlab.agent.command+json;version=1"
    AGENT_STATUS = "application/vnd.imecilabt.solidlab.agent.status+json;version=1"
    AGENT_COMMAND_RESULT = (
        "application/vnd.imecilabt.solidlab.agent.commandres+json;version=1"
    )

    TEXT = "text/plain"
    JSON = "application/json"

    # def from_type(self, t: Type) -> 'SolidLabContentType':
    #     pass  # hard to implement with support for subclasses, as "t is AgentCommand" ignores subclasses
    #  (can be implemented anyway, when actually needed)

    def matches_type(self, obj: Any) -> bool:
        if isinstance(obj, AgentCommand):
            return self == SolidLabContentType.AGENT_COMMAND
        if isinstance(obj, Agent):
            return self == SolidLabContentType.AGENT_STATUS
        if isinstance(obj, AgentCommandResult):
            return self == SolidLabContentType.AGENT_COMMAND_RESULT
        raise ValueError(f"obj {obj!r} of type {type(obj)!r} does not match {self!r}")



import json
import logging
import mimetypes
from typing import Tuple, Optional

from urllib3 import Timeout


def validate_artifact_endpoint(value: str) -> str:
    """
    Validates an artifact endpoint. Accepts multiple partial endpoints.
    Always returns a canonical value, ending in: /artifact/
    :param value: an artifact endpoint, optionally including the end /artifact/
    :return: a canonical value, ending in: /artifact/
    """
    # TODO use full regex to check if only allowed forms are given
    #        https?://.*/(experiment|perftest|testenv)/[0-9]+/?(artifact)?/?
    if not isinstance(value, str):
        raise ValueError(
            f"artifact_endpoint must be a URL (and thus a string and not {type(value)})"
        )
    if len(value.strip()) == 0:
        raise ValueError(f"artifact_endpoint must not be empty ({value!r})")
    if not value.startswith("https://") and not value.startswith("http://"):
        raise ValueError(f"artifact_endpoint must be a URL, not {value!r}")
    if (
        "/perftest/" not in value
        and "/testenv/" not in value
        and "/experiment/" not in value
    ):
        raise ValueError(
            "artifact_endpoint must contain a full object endpoint "
            "(and thus contain /perftest/{id}/ /experiment/{id}/ or /testenv/{id}/)"
        )
    if not value.endswith("/"):
        value = value + "/"
    if (
        value.endswith("perftest/")
        or value.endswith("experiment/")
        or value.endswith("testenv/")
    ):
        raise ValueError(
            "artifact_endpoint must contain a full object endpoint "
            "(and thus contain /perftest/{id}/ /experiment/{id}/ or /testenv/{id}/)"
        )
    if not value.endswith("/artifact/"):
        value = value + "artifact/"
    return value[:-1]


def upload_artifact_file(
    session,
    *,
    artifact_endpoint: str,
    attach_type: str,
    sub_type: str,
    description: str,
    filename: str,
    content_type: Optional[str] = None,
    auth_token: Optional[str] = None,
) -> Tuple[int, str]:
    if not content_type:
        guesses = mimetypes.guess_type(filename)
        if not guesses:
            raise ValueError(f"mime type could not be guessed from filename.")
        content_type = guesses[0]
        # logger.debug(f"Guessed mime-type from filename: {mime_type!r}")

    with open(filename, "rb") as f:
        content = f.read()

        return upload_artifact(
            session,
            artifact_endpoint=artifact_endpoint,
            attach_type=attach_type,
            sub_type=sub_type,
            description=description,
            content=content,
            content_type=content_type,
            auth_token=auth_token,
        )


def upload_artifact(
    session,
    *,
    artifact_endpoint: str,
    attach_type: str,
    sub_type: str,
    description: str,
    content_type: str,
    content: bytes,
    auth_token: Optional[str] = None,
) -> Tuple[int, str]:
    """

    :param session:
    :param artifact_endpoint:
    :param attach_type:
    :param sub_type:
    :param description:
    :param content_type: examples "text/csv" "text/plain" "image/svg+xml" "image/png"
    :param content:
    :return:
    """
    artifact_endpoint = validate_artifact_endpoint(artifact_endpoint)

    headers = {
        "Content-Type": content_type,
        "X-Solidlab-Artifact-Type": attach_type,
        "X-Solidlab-Artifact-Subtype": sub_type,
        "X-Solidlab-Artifact-Description": description,
    }

    if auth_token:
        # Solidlab-Perftest-Auth is also supported
        headers["X-Solidlab-Perftest-Auth"] = auth_token

    post_attach_meta_resp = session.post(
        artifact_endpoint,
        params={},
        headers=headers,
        timeout=Timeout(connect=2.0, read=3.0),
        data=content,
    )
    post_attach_meta_resp.raise_for_status()
    resp_json = post_attach_meta_resp.json()
    if "@id" not in resp_json or "id" not in resp_json:
        logging.error(
            f"Unexpected reply when POSTing artifact to {artifact_endpoint}: \n{json.dumps(resp_json, indent=3)}"
        )
    artifact_url = resp_json["@id"]
    artifact_id = resp_json["id"]
    return artifact_id, artifact_url



from typing import List, TypeVar, Generic

from pydantic import BaseModel, PositiveInt, NonNegativeInt
from pydantic.generics import GenericModel


class PageRequest(BaseModel):
    # Index of requested page (first page is 1!)
    page: PositiveInt

    # Max items per page
    size: PositiveInt


ItemT = TypeVar("ItemT")


class Page(GenericModel, Generic[ItemT]):
    # The items on this page. len(items) == 0 if total is 0. len(items) can be smaller than size.
    items: List[ItemT]

    # Index of this returned page (first page is 1!)
    page: PositiveInt

    # Max items per page (so can be more than actual number of items on last page)
    size: PositiveInt

    # Total items, all pages combined
    total: NonNegativeInt



import asyncio
import logging
import traceback
from asyncio import StreamReader, FIRST_EXCEPTION
from datetime import datetime, timezone, timedelta
from subprocess import PIPE, STDOUT, Popen
from typing import List, Tuple, Optional, Dict, TypeAlias, Callable

from dateutil.parser import parse as dateutil_parse


def dump_rfc3339(dt: datetime, *, no_milliseconds=True) -> str:
    assert dt.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    if no_milliseconds:
        dt = dt.replace(microsecond=0)  # not interested in sub second precision
    res = dt.isoformat()
    if res.endswith("+00:00"):
        # +00:00 is nicer to write as Z (Zulu timezone)
        # But both are valid.
        res = res[:-6] + "Z"
    return res


def parse_rfc3339(date_str: str) -> datetime:
    # dateutil_parse parses ISO 8601, which contains almost all of RFC3339. This is good enough.
    res = dateutil_parse(date_str)
    assert res.tzinfo is not None  # assert timezone aware (non-naive) datetime
    return res


def datetime_now(*, zulu=False, no_milliseconds=False) -> datetime:
    res = datetime.now(timezone.utc)
    if not zulu:
        res = res.astimezone(timezone(timedelta(hours=0), name="UTC"))
    if no_milliseconds:
        res = res.replace(microsecond=0)
    assert res.tzinfo is not None  # enforce timezone aware (non-naive) datetime
    return res


def call_external_command_sync(*params: str, **kwargs) -> Tuple[int, str]:
    res = ""
    com: List[str] = list(params)
    kwargs["stdout"] = PIPE
    kwargs["stderr"] = STDOUT
    kwargs["universal_newlines"] = True
    with Popen(com, **kwargs) as proc:
        for line in proc.stdout:
            res += line
        res_code = proc.wait()
        # if res_code != 0:
        #     raise Exception('Command returned error ({}): {} (command={} )'.format(res_code, res.strip(), com))
        return res_code, res.strip()


OutputHandler: TypeAlias = Callable[[str], None]
NO_OUTPUT = lambda msg: None


async def log_output(out: OutputHandler, stream: StreamReader):
    saw_eof = False
    while not saw_eof:
        line = await stream.readline()
        if line:
            out(line.decode())
        else:
            saw_eof = True


async def call_external_command_async(
    *,
    command: List[str],
    cwd: Optional[str] = None,
    env: Dict[str, str] = {},
    timeout_s=3600,
    stdout_handler: OutputHandler = NO_OUTPUT,
    stderr_handler: OutputHandler = NO_OUTPUT,
    debug_log_handler: OutputHandler = NO_OUTPUT,
    error_log_handler: OutputHandler = NO_OUTPUT,
) -> int:
    assert command
    assert isinstance(command, list)
    assert all(isinstance(c, str) for c in command)

    env = dict(env)
    # await add_log(
    #     perftest_id,
    #     LogLevel.DEBUG,
    #     f"Will run command: {command!r} with env={env} in cwd={cwd}",
    #     session=session,
    # )
    proc = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=cwd,
        env=env,
    )
    tasks = []
    stdout_log_task = asyncio.create_task(log_output(stdout_handler, proc.stdout))
    stderr_log_task = asyncio.create_task(log_output(stderr_handler, proc.stderr))
    tasks.append(stdout_log_task)
    tasks.append(stderr_log_task)
    proc_done_task = asyncio.create_task(proc.wait())
    tasks.append(proc_done_task)

    pre_wait = datetime_now()

    debug_log_handler(
        f"Started command. Will wait for up to {timeout_s} seconds. Starting at {pre_wait}"
    )

    done, pending = await asyncio.wait(
        tasks,
        timeout=timeout_s,
        return_when=FIRST_EXCEPTION,
    )
    post_wait = datetime_now()
    logging.debug(
        f"(pre_wait={pre_wait} post_wait={post_wait}) len(done)={len(done)} len(pending)={len(pending)}"
    )
    if pending:
        saw_exception = False
        for d in done:
            e = d.exception()
            if e:
                saw_exception = True
                logging.error("Saw exception waiting", exc_info=e)
                tb = "".join(traceback.format_exception(None, e, e.__traceback__))
                error_log_handler(
                    f"Command exception: Will be cancelled. "
                    f"(pre_wait={pre_wait} post_wait={post_wait} e={e} {tb})",
                )
        if not saw_exception:
            error_log_handler(
                f"Command did not return within {timeout_s} seconds: Will be cancelled. "
                f"(pre_wait={pre_wait} post_wait={post_wait})",
            )
        for p in pending:
            p.cancel()
            proc.kill()
        return -1
    else:
        debug_log_handler(
            f"Command exited with exitcode {proc.returncode}",
        )
        return proc.returncode


def is_naive_dt(v: datetime) -> bool:
    return v.tzinfo is None


def is_non_naive_dt(v: datetime) -> bool:
    return v.tzinfo is not None


def is_non_naive_opt_dt(v: Optional[datetime]) -> bool:
    return v is None or v.tzinfo is not None


def assert_non_naive_dt(v: datetime):
    assert v.tzinfo is not None, 'time is naive "{}"'.format(v)


def assert_non_naive_opt_dt(v: Optional[datetime]):
    assert v is None or v.tzinfo is not None, 'time is naive "{}"'.format(v)


def count_none(*args) -> int:
    """
    :return: the number of arguments that is None
    """
    res = 0
    for arg in args:
        res += int(arg is None)
    return res


def count_not_none(*args) -> int:
    """
    :return: the number of arguments that is not None
    """
    res = 0
    for arg in args:
        res += int(arg is not None)
    return res


def is_single_not_none(*args) -> bool:
    """
    :return: True if exactly one of the arguments is not None
    """
    res = 0
    for arg in args:
        res += int(arg is not None)
    return res == 1


def is_single_none(*args) -> bool:
    """
    :return: True if exactly one of the arguments is None
    """
    res = 0
    for arg in args:
        res += int(arg is None)
    return res == 1



import enum
from datetime import timedelta, datetime
from typing import Optional, Union, List, Dict, Any

from pydantic import BaseModel


class AgentCommandType(enum.Enum):
    WRITE_FILE = "WRITE_FILE"
    EXEC_BASH = "EXEC_BASH"
    START_SERVICE = "START_SERVICE"
    STOP_SERVICE = "STOP_SERVICE"


class AgentCommandResultStatus(enum.Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"
    TIMEOUT = "TIMEOUT"


class Agent(BaseModel):
    testenv_id: int
    machine_id: str
    last_active: datetime
    active: bool


class AgentCommandResult(BaseModel):
    command_id: int

    status: AgentCommandResultStatus
    started: datetime
    stopped: datetime

    # output of command, all optional
    return_value: Optional[int] = None
    error_msg: Optional[str] = None
    trace: Optional[str] = None
    debug_msg: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class AgentCommand(BaseModel):
    id: int
    testenv_id: int  # testenv_id + machine_id identifies agent the command belongs too
    machine_id: str  # testenv_id + machine_id identifies agent the command belongs too
    experiment_id: Optional[int]
    created: datetime
    name: str  # hrn of the command, ignored by agent, but useful to inform users
    type: AgentCommandType
    # data:
    #   CSS config filename, or bash commands array, or service name. Type depends on AgentCommandType.
    #   (Dict[str, Any] is not used yet, but might be in the future)
    #
    # data type by AgentCommandType:
    #   - WRITE_FILE -> List[str] with 2 elements ([filename, file_content])
    #   - EXEC_BASH -> List[str] of command parts
    #   - START_SERVICE -> str (service name)
    #   - STOP_SERVICE -> str (service name)
    #
    data: Union[str, List[str], Dict[str, Any]]
    timeout_s: float
    started: Optional[datetime]  # only set when command is started
    # implicit: deadline = started + timeout_s
    # implicit: is_expired = no AgentCommandResult and now > deadline


class AgentCommandFull(AgentCommand):
    result: Optional[AgentCommandResult]
